package org.japura.controller;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.japura.MethodNotImplementedException;

/**
 * <P>
 * Copyright (C) 2013 Carlos Eduardo Leite de Andrade
 * <P>
 * This library is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * <P>
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * <P>
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <A
 * HREF="www.gnu.org/licenses/">www.gnu.org/licenses/</A>
 * <P>
 * For more information, contact: <A HREF="www.japura.org">www.japura.org</A>
 * <P>
 * 
 * @author Carlos Eduardo Leite de Andrade
 */
abstract class AbstractControllerManager implements ControllerManager{

  ArrayList<ControllerModel> controllers;
  List<ControllerManagerListener> listeners;

  private List<Link> links;
  private ControllerMockFactory controllerMockFactory;

  AbstractControllerManager() {
	controllers = new ArrayList<ControllerModel>();
	links = new ArrayList<Link>();
	listeners = new ArrayList<ControllerManagerListener>();
  }

  ControllerMockFactory getControllerMockFactory() {
	return controllerMockFactory;
  }

  void setControllerMockFactory(ControllerMockFactory controllerMockFactory) {
	this.controllerMockFactory = controllerMockFactory;
  }

  public final void addListener(ControllerManagerListener listener) {
	synchronized (listeners) {
	  listeners.add(listener);
	}
  }

  public final void removeListener(ControllerManagerListener listener) {
	synchronized (listeners) {
	  listeners.remove(listener);
	}
  }

  public final List<ControllerManagerListener> getListeners() {
	synchronized (listeners) {
	  return Collections.unmodifiableList(listeners);
	}
  }

  final void methodNotImplemented() {
	throw new MethodNotImplementedException();
  }

  final void testControllerModel(Class<?> clss) {
	if (ControllerModel.class.isAssignableFrom(clss) == false) {
	  throw new InvalidControllerClassException(clss);
	}
  }

  @Override
  public final boolean existsLink(ControllerModel controller1,
								  ControllerModel controller2) {
	if (controller1 == null || controller2 == null) {
	  return false;
	}

	for (Link link : links) {
	  if (link.contains(controller1) && link.contains(controller2)) {
		return true;
	  }
	}
	return false;
  }

  @Override
  public final boolean existsLink(ControllerModel controller, Class<?> clss) {
	testControllerModel(clss);

	if (controller == null || clss == null
		|| controller.getClass().equals(clss)) {
	  return false;
	}

	for (Link link : links) {
	  if (link.contains(controller) && link.contains(clss)) {
		return true;
	  }
	}
	return false;
  }

  @Override
  public final void link(ControllerModel controller1,
						 ControllerModel controller2) {
	if (controller1 == null || controller2 == null) {
	  throw new ControllerException("Null controller");
	}
	if (controller1.getClass().equals(controller2.getClass())) {
	  throw new ControllerException(
		  "Can not link two controllers from the same class");
	}
	if (controller1.equals(controller2)) {
	  throw new ControllerException("Can not link the same controller");
	}

	for (Link link : links) {
	  if (link.contains(controller1)) {
		throw new ControllerException("Already exists a link for "
			+ controller1.getClass().getName() + " ["
			+ controller1.getControllerId() + "]");
	  }
	  if (link.contains(controller2)) {
		throw new ControllerException("Already exists a link for "
			+ controller2.getClass().getName() + " ["
			+ controller2.getControllerId() + "]");
	  }
	}
	links.add(new Link(controller1, controller2));
  }

  @Override
  public final void unlink(ControllerModel controller1,
						   ControllerModel controller2) {
	if (controller1 == null || controller2 == null) {
	  return;
	}
	if (controller1.getClass().equals(controller2.getClass())) {
	  return;
	}
	if (controller1.equals(controller2)) {
	  return;
	}
	Link link = null;
	for (Link l : links) {
	  if (l.contains(controller1) && l.contains(controller2)) {
		link = l;
		break;
	  }
	}
	if (link != null) {
	  links.remove(link);
	}
  }

  @Override
  public final void remove(String id) {
	remove(id, true);
  }

  @Override
  public final void remove(String id, boolean cancelTaskExecution) {
	if (id != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getControllerId().equals(id)) {
		  remove(controller, cancelTaskExecution);
		  break;
		}
	  }
	}
  }

  @Override
  public final void removeGroup(String groupdId) {
	List<String> ids = new ArrayList<String>();
	for (ControllerModel controller : controllers) {
	  if (controller.getGroupId().equals(groupdId)) {
		ids.add(controller.getControllerId());
	  }
	}
	for (String id : ids) {
	  remove(id);
	}
  }

  @Override
  public final void removeAll(Class<?> clss) {
	testControllerModel(clss);
	synchronized (listeners) {
	  for (ControllerManagerListener listener : listeners) {
		listener.attemptToRemove(clss);
	  }
	}
	List<String> ids = new ArrayList<String>();
	for (ControllerModel controller : controllers) {
	  if (clss.isAssignableFrom(controller.getClass())) {
		ids.add(controller.getControllerId());
	  }
	}
	for (String id : ids) {
	  remove(id);
	}
  }

  @Override
  public final void removeAllFromGroup(String groupId, Class<?> clss) {
	testControllerModel(clss);
	if (groupId != null) {
	  synchronized (listeners) {
		for (ControllerManagerListener listener : listeners) {
		  listener.attemptToRemove(groupId, clss);
		}
	  }
	  List<String> ids = new ArrayList<String>();
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)
			&& clss.isAssignableFrom(controller.getClass())) {
		  ids.add(controller.getControllerId());
		}
	  }
	  for (String id : ids) {
		remove(id);
	  }
	}
  }

  @Override
  public final boolean isInstancied(Class<?> clss) {
	testControllerModel(clss);
	for (ControllerModel controller : controllers) {
	  if (clss.isAssignableFrom(controller.getClass())) {
		return true;
	  }
	}
	return false;
  }

  @Override
  public final boolean isInstanciedInGroup(String groupId, Class<?> clss) {
	testControllerModel(clss);
	if (groupId != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)
			&& clss.isAssignableFrom(controller.getClass())) {
		  return true;
		}
	  }
	}
	return false;
  }

  @Override
  public final void removeAll() {
	List<String> ids = new ArrayList<String>();
	for (ControllerModel controller : controllers) {
	  ids.add(controller.getControllerId());
	}
	for (String id : ids) {
	  remove(id);
	}
  }

  @Override
  public final int count(Class<?> clss) {
	testControllerModel(clss);

	int count = 0;
	for (ControllerModel controller : controllers) {
	  if (clss.isAssignableFrom(controller.getClass())) {
		count++;
	  }
	}
	return count;
  }

  @Override
  public final <E> E get(Class<E> clss) {
	testControllerModel(clss);

	ChildController cc = clss.getAnnotation(ChildController.class);
	if (cc != null && cc.getOnlyFromGroup()) {
	  throw new ControllerException(
		  "You need get the child controller "
			  + clss.getName()
			  + " through the group. Use getGroup().get() or Controller.getFromGroup() method.");
	}

	for (ControllerModel controller : controllers) {
	  if (clss.isAssignableFrom(controller.getClass())) {
		return clss.cast(controller);
	  }
	}
	return newInstance(clss);
  }

  @Override
  public final <E> E newInstance(Class<E> clss) {
	E controller = null;
	if (getNewInstanceClass().isAssignableFrom(clss)) {
	  try {
		if (getControllerMockFactory() != null) {
		  controller = getControllerMockFactory().create(clss);
		} else {
		  controller = clss.newInstance();
		}
	  } catch (Exception e) {
		e.printStackTrace();
		AbstractController.newParentController = null;
		throw new RuntimeException(e);
	  }
	} else {
	  AbstractController.newParentController = null;
	  throw new InvalidControllerClassException(clss);
	}
	return controller;
  }

  @Override
  public final <E> E get(String id, Class<E> clss) {
	if (id != null) {
	  testControllerModel(clss);
	  for (ControllerModel controller : controllers) {
		if (controller.getControllerId().equals(id)
			&& clss.isAssignableFrom(controller.getClass())) {
		  return clss.cast(controller);
		}
	  }
	}
	return null;
  }

  @Override
  public final <E> E getFromGroup(String groupId, Class<E> clss) {
	testControllerModel(clss);
	for (ControllerModel controller : controllers) {
	  if (controller.getGroupId().equals(groupId)
		  && clss.isAssignableFrom(controller.getClass())) {
		return clss.cast(controller);
	  }
	}
	return null;
  }

  @Override
  public final <E> E getFromGroup(String groupId, String id, Class<E> clss) {
	testControllerModel(clss);
	if (groupId != null && id != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)
			&& controller.getControllerId().equals(id)
			&& clss.isAssignableFrom(controller.getClass())) {
		  return clss.cast(controller);
		}
	  }
	}
	return null;
  }

  @Override
  public final ControllerModel getFromGroup(String groupId, String id) {
	if (groupId != null && id != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)
			&& controller.getControllerId().equals(id)) {
		  return controller;
		}
	  }
	}
	return null;
  }

  @Override
  public final <E> List<E> getAllFromGroup(String groupId, Class<E> clss) {
	testControllerModel(clss);
	List<E> list = new ArrayList<E>();
	if (groupId != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)
			&& clss.isAssignableFrom(controller.getClass())) {
		  list.add(clss.cast(controller));
		}
	  }
	}
	return list;
  }

  @Override
  public final List<ControllerModel> getAllFromGroup(String groupId) {
	List<ControllerModel> list = new ArrayList<ControllerModel>();
	if (groupId != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)) {
		  list.add(controller);
		}
	  }
	}
	return list;
  }

  @Override
  public final ControllerModel get(String id) {
	if (id != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getControllerId().equals(id)) {
		  return controller;
		}
	  }
	}
	return null;
  }

  @Override
  public final <E> List<E> getAll(Class<E> clss) {
	testControllerModel(clss);
	List<E> list = new ArrayList<E>();
	for (ControllerModel controller : controllers) {
	  if (clss.isAssignableFrom(controller.getClass())) {
		list.add(clss.cast(controller));
	  }
	}
	return list;
  }

  @Override
  public final List<ControllerModel> getAll() {
	return new ArrayList<ControllerModel>(controllers);
  }

  @Override
  public final boolean contains(ControllerModel controller) {
	if (controller.isRemoved()) {
	  return false;
	}
	return controllers.contains(controller);
  }

  @Override
  public final void printAllControllers() {
	for (ControllerModel controller : controllers) {
	  System.out.println("Id: " + controller.getControllerId() + " GroupId: "
		  + controller.getGroupId() + " Class: "
		  + controller.getClass().getName());
	}
  }

  @Override
  public final ControllerModel getRoot(String groupId) {
	if (groupId != null) {
	  for (ControllerModel controller : controllers) {
		if (controller.getGroupId().equals(groupId)) {
		  return controller.getRoot();
		}
	  }
	}
	return null;
  }

  @Override
  public final <E> E getFromLink(ControllerModel controllerInLink, Class<E> clss) {
	testControllerModel(clss);
	if (controllerInLink == null || clss == null) {
	  return null;
	}
	if (controllerInLink.getClass().equals(clss)) {
	  return null;
	}

	for (Link link : links) {
	  if (link.contains(controllerInLink)) {
		return link.get(clss);
	  }
	}
	return null;
  }

  @Override
  public final void unlink(ControllerModel controller) {
	Link link = null;
	for (Link l : links) {
	  if (l.contains(controller)) {
		link = l;
		break;
	  }
	}
	if (link != null) {
	  links.remove(link);
	}
  }

  @Override
  public final void fetchHierarchy(ControllerModel parentController,
								   List<ControllerModel> list) {
	List<ControllerModel> children = parentController.getChildren();
	for (ControllerModel child : children) {
	  ControllerModel c = (ControllerModel) child;
	  list.add(c);
	  fetchHierarchy(c, list);
	}
  }

  @Override
  public final void add(ControllerModel controller) {
	controllers.add(controller);
	synchronized (listeners) {
	  for (ControllerManagerListener listener : listeners) {
		listener.added(controller);
	  }
	}
  }

  @Override
  public final void publishToAll(Message message) {
	for (ControllerModel controller : getAll()) {
	  if (message.isConsumed()) {
		break;
	  }
	  if (message.getPublisher() != null
		  && message.isIgnorePublisher()
		  && message.getPublisher().getControllerId().equals(controller
			  .getControllerId())) {
		continue;
	  }
	  if (acceptsMessage(controller, message)
		  && message.acceptsController(controller)) {
		controller.subscribe(message);
	  }
	}
  }

  @Override
  public boolean acceptsMessage(ControllerModel controller, Message message) {
	for (MessageFilter filter : controller.getMessageFilters()) {
	  if (filter.accepts(message) == false) {
		return false;
	  }
	}
	return true;
  }

}
