package org.japura.task;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.locks.ReentrantLock;

import javax.swing.SwingUtilities;

import org.japura.controller.ControllerModel;
import org.japura.controller.Group;
import org.japura.debug.DebugComponent;
import org.japura.debug.DebugResult;
import org.japura.debug.DebugWindow;
import org.japura.task.DebugTasks.AddParameter;
import org.japura.task.DebugTasks.AddParameterEvent;
import org.japura.task.DebugTasks.RemoveParameter;
import org.japura.task.ui.DefaultTaskExecutionUI;

/**
 * <P>
 * Copyright (C) 2011-2012 Carlos Eduardo Leite de Andrade
 * <P>
 * This library is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * <P>
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * <P>
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <A
 * HREF="www.gnu.org/licenses/">www.gnu.org/licenses/</A>
 * <P>
 * For more information, contact: <A HREF="www.japura.org">www.japura.org</A>
 * <P>
 * 
 * @author Carlos Eduardo Leite de Andrade
 */
public class DefaultTaskManager extends AbstractTaskManager{

  private ReentrantLock lock = new ReentrantLock();

  /** for tasks without controller */
  private SerialTaskExecutor serialExecutor = new SerialTaskExecutor();
  /** for any task with asynchronous mode */
  private ParallelTaskExecutor parallelExecutor = new ParallelTaskExecutor();
  /** for tasks from controller group */
  private Map<Group, SerialTaskExecutor> groupExecutors =
	  new HashMap<Group, SerialTaskExecutor>();

  public DefaultTaskManager() {
	setTaskExecutionUI(new DefaultTaskExecutionUI());
  }

  @Override
  public List<DebugResult> addToDebugWindow(Task<?> task, TaskEvent event) {
	if (DebugComponent.exists(DebugTasks.class)) {
	  AddParameterEvent parameter =
		  new AddParameterEvent(task, event, System.currentTimeMillis(),
			  SwingUtilities.isEventDispatchThread(), getQueueCount());
	  return DebugComponent.publish(DebugTasks.class, parameter);
	}
	return null;
  }

  @Override
  public void addToDebugWindow(Task<?> task) {
	if (DebugComponent.exists(DebugTasks.class)) {
	  AddParameter parameter = new AddParameter(task);
	  DebugComponent.publish(DebugTasks.class, parameter);
	}
  }

  @Override
  public void removeFromDebugWindow(Task<?> task) {
	if (DebugComponent.exists(DebugTasks.class)) {
	  RemoveParameter parameter = new RemoveParameter(task);
	  DebugComponent.publish(DebugTasks.class, parameter);
	}
  }

  @Override
  public void submit(Task<?> task, TaskSession session, boolean isParallel) {
	lock.lock();
	try {
	  if (task.isSubmitted()) {
		throw new TaskExeception("Task already submitted");
	  }
	  task.setSubmitted(true);

	  session = getTaskSessionFactory().buildTaskSession(task, session);
	  checkSession(task, session);
	  task.setSession(session);

	  task.setParallel(isParallel);

	  addToDebugWindow(task, TaskEvent.SUBMIT);

	  task.submitted();

	  fireTaskManagerListeners(TaskManagerEvent.SUBMITTED, task);
	  fireTaskExecutionUIs(TaskManagerEvent.SUBMITTED, task);

	  if (isParallel) {
		task.setCancelToken(parallelExecutor.getCancelToken());
		parallelExecutor.submit(task);
		return;
	  }

	  SerialTaskExecutor executor = serialExecutor;
	  ControllerModel controller = task.getRootModel();
	  if (controller != null) {
		Group group = controller.getGroup();
		executor = groupExecutors.get(group);
		if (executor == null) {
		  executor = new SerialTaskExecutor();
		  groupExecutors.put(group, executor);
		}
	  }

	  if (executor == null) {
		throw new TaskExeception("There is no executor");
	  }

	  task.setCancelToken(executor.getCancelToken());

	  executor.submit(task);
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public boolean hasTask(Group group) {
	if (parallelExecutor.hasTask(group.getId())) {
	  return true;
	}
	lock.lock();
	try {
	  SerialTaskExecutor executor = groupExecutors.get(group);
	  if (executor != null) {
		return executor.hasTask();
	  }
	} finally {
	  lock.unlock();
	}
	return false;
  }

  @Override
  public boolean hasTask() {
	lock.lock();
	try {
	  for (SerialTaskExecutor executor : groupExecutors.values()) {
		if (executor.hasTask()) {
		  return true;
		}
	  }
	} finally {
	  lock.unlock();
	}
	return serialExecutor.hasTask() || parallelExecutor.hasTask();
  }

  @Override
  public void cancel(String controllerGroupId) {
	if (controllerGroupId != null) {
	  lock.lock();
	  try {
		for (Entry<Group, SerialTaskExecutor> entry : groupExecutors.entrySet()) {
		  Group group = entry.getKey();
		  if (group.getId().equals(controllerGroupId)) {
			entry.getValue().cancel();
		  }
		}
	  } finally {
		lock.unlock();
	  }
	}
  }

  @Override
  public void cancelAll() {
	parallelExecutor.cancel();
	serialExecutor.cancel();
	lock.lock();
	try {
	  for (SerialTaskExecutor executor : groupExecutors.values()) {
		executor.cancel();
	  }
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public int getQueueCount() {
	int count = 0;
	lock.lock();
	try {
	  for (SerialTaskExecutor executor : groupExecutors.values()) {
		count += executor.getQueue().size();
	  }
	} finally {
	  lock.unlock();
	}
	count += serialExecutor.getQueue().size();
	count += parallelExecutor.getQueue().size();
	return count;
  }

  @Override
  public DebugComponent createDebugPanel() {
	DebugTasks component = new DebugTasks();
	DebugComponent.add(component);
	return component;
  }

  @Override
  public void showDebugWindow() {
	DebugWindow.showDebugWindow(createDebugPanel());
  }

  @Override
  public void notifyAfterTaskExecution(Task<?> task) {
	lock.lock();
	try {
	  ControllerModel controller = task.getRootModel();
	  if (controller == null) {
		return;
	  }
	  Group group = controller.getGroup();
	  SerialTaskExecutor executor = groupExecutors.get(group);
	  if (executor != null && executor.hasTask() == false) {
		groupExecutors.remove(group);
	  }
	} finally {
	  lock.unlock();
	  addToDebugWindow(task, TaskEvent.AFTER);
	  removeFromDebugWindow(task);
	  fireTaskManagerListeners(TaskManagerEvent.AFTER_EXECUTE, task);
	  fireTaskExecutionUIs(TaskManagerEvent.AFTER_EXECUTE, task);
	}
  }
}
