package org.japura.task;

import java.util.AbstractQueue;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * <P>
 * Copyright (C) 2013 Carlos Eduardo Leite de Andrade
 * <P>
 * This library is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * <P>
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * <P>
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <A
 * HREF="www.gnu.org/licenses/">www.gnu.org/licenses/</A>
 * <P>
 * For more information, contact: <A HREF="www.japura.org">www.japura.org</A>
 * <P>
 * 
 * @author Carlos Eduardo Leite de Andrade
 */
class SerialBlockingQueue extends AbstractQueue<Runnable> implements
	BlockingQueue<Runnable>{

  private final LinkedList<Runnable> queue;
  private final ReentrantLock lock = new ReentrantLock(true);
  private final Condition notEmpty = lock.newCondition();

  public SerialBlockingQueue() {
	queue = new LinkedList<Runnable>();
  }

  @Override
  public boolean add(Runnable e) {
	return offer(e);
  }

  @Override
  public boolean offer(Runnable e) {
	lock.lock();
	try {
	  FutureTaskWrapper<?> fw = (FutureTaskWrapper<?>) e;
	  if (fw.getTask().isPriority()) {
		queue.add(0, e);
		fw.getTask().clearPriority();
	  } else {
		queue.add(e);
	  }
	  notEmpty.signal();
	  return true;
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public void put(Runnable e) {
	offer(e);
  }

  @Override
  public boolean offer(Runnable e, long timeout, TimeUnit unit) {
	return offer(e);
  }

  @Override
  public Runnable take() throws InterruptedException {
	lock.lockInterruptibly();
	try {
	  try {
		while (queue.size() == 0) {
		  notEmpty.await();
		}
	  } catch (InterruptedException ie) {
		notEmpty.signal();
		throw ie;
	  }
	  Runnable x = queue.poll();
	  return x;
	} finally {
	  lock.unlock();
	}
  }

  public Runnable poll() {
	lock.lock();
	try {
	  return queue.poll();
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public Runnable poll(long timeout, TimeUnit unit) throws InterruptedException {
	long nanos = unit.toNanos(timeout);
	lock.lockInterruptibly();
	try {
	  for (;;) {
		Runnable x = queue.poll();
		if (x != null) {
		  return x;
		}
		if (nanos <= 0) {
		  return null;
		}
		try {
		  nanos = notEmpty.awaitNanos(nanos);
		} catch (InterruptedException ie) {
		  notEmpty.signal();
		  throw ie;
		}
	  }
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public Runnable peek() {
	lock.lock();
	try {
	  return queue.peek();
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public int size() {
	lock.lock();
	try {
	  return queue.size();
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public int remainingCapacity() {
	return Integer.MAX_VALUE;
  }

  @Override
  public boolean remove(Object o) {
	lock.lock();
	try {
	  return queue.remove(o);
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public boolean contains(Object o) {
	lock.lock();
	try {
	  return queue.contains(o);
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public Object[] toArray() {
	lock.lock();
	try {
	  return queue.toArray();
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public String toString() {
	lock.lock();
	try {
	  return queue.toString();
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public int drainTo(Collection<? super Runnable> c) {
	if (c == null)
	  throw new NullPointerException();
	if (c == this)
	  throw new IllegalArgumentException();
	final ReentrantLock lock = this.lock;
	lock.lock();
	try {
	  int n = 0;
	  Runnable e;
	  while ((e = queue.poll()) != null) {
		c.add(e);
		++n;
	  }
	  return n;
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public int drainTo(Collection<? super Runnable> c, int maxElements) {
	if (c == null) {
	  throw new NullPointerException();
	}
	if (c == this) {
	  throw new IllegalArgumentException();
	}
	if (maxElements <= 0) {
	  return 0;
	}
	lock.lock();
	try {
	  int n = 0;
	  Runnable e;
	  while (n < maxElements && (e = queue.poll()) != null) {
		c.add(e);
		++n;
	  }
	  return n;
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public void clear() {
	lock.lock();
	try {
	  queue.clear();
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public <T> T[] toArray(T[] a) {
	lock.lock();
	try {
	  return queue.toArray(a);
	} finally {
	  lock.unlock();
	}
  }

  @Override
  public Iterator<Runnable> iterator() {
	return new Itr(toArray());
  }

  private class Itr implements Iterator<Runnable>{
	final Object[] array;
	int cursor;
	int lastRet;

	Itr(Object[] array) {
	  lastRet = -1;
	  this.array = array;
	}

	public boolean hasNext() {
	  return cursor < array.length;
	}

	public Runnable next() {
	  if (cursor >= array.length)
		throw new NoSuchElementException();
	  lastRet = cursor;
	  return (Runnable) array[cursor++];
	}

	public void remove() {
	  if (lastRet < 0)
		throw new IllegalStateException();
	  Object x = array[lastRet];
	  lastRet = -1;
	  lock.lock();
	  try {
		for (Iterator<?> it = queue.iterator(); it.hasNext();) {
		  if (it.next() == x) {
			it.remove();
			return;
		  }
		}
	  } finally {
		lock.unlock();
	  }
	}
  }

}
